
{
    module P = Parse
    
    exception Scan of string
    
    let error msg   = raise (Scan msg)
    
    let get         = Lexing.lexeme
    let getchar     = Lexing.lexeme_char

    type mode = Open | Close
}

let ws          = [' ' '\n' '\r' '\t']
let printable   = ['!'-';' '=' '?'-'~'] (* all printable, except '<','>' *) 

rule token = parse      
    eof         { P.EOF }
  | "<"         { tag lexbuf Open   }
  | "</"        { tag lexbuf Close  } 
  | ws+         { P.SPACE(get lexbuf) } 
  | printable+  { P.TEXT(get lexbuf)  }
  | _           { error ("illegal character: '"^get lexbuf^"'") }   

and tag = parse
  | "B>"        { function Open -> P.OB  | Close -> P.CB  }
  | "EM>"       { function Open -> P.OEM | Close -> P.CEM }
  | "I>"        { function Open -> P.OI  | Close -> P.CI  }
  | "PL>"       { function Open -> P.OPL | Close -> P.CPL }
  | "S>"        { function Open -> P.OS  | Close -> P.CS  }
  | "TT>"       { function Open -> P.OTT | Close -> P.CTT }
  | "U>"        { function Open -> P.OU  | Close -> P.CU  }
  | "0>"        { function Open -> P.OS0 | Close -> P.CS0 }
  | "1>"        { function Open -> P.OS1 | Close -> P.CS1 }
  | "2>"        { function Open -> P.OS2 | Close -> P.CS2 }
  | "3>"        { function Open -> P.OS3 | Close -> P.CS3 }
  | "4>"        { function Open -> P.OS4 | Close -> P.CS4 }
  | "5>"        { function Open -> P.OS5 | Close -> P.CS5 }
  | "6>"        { function Open -> P.OS6 | Close -> P.CS6 }
  | "7>"        { function Open -> P.OS7 | Close -> P.CS7 }
  | "8>"        { function Open -> P.OS8 | Close -> P.CS8 }
  | "9>"        { function Open -> P.OS9 | Close -> P.CS9 }
  | "r>"        { function Open -> P.OCR | Close -> P.CCR }
  | "g>"        { function Open -> P.OCG | Close -> P.CCG }
  | "b>"        { function Open -> P.OCB | Close -> P.CCB }
  | "c>"        { function Open -> P.OCC | Close -> P.CCC }
  | "m>"        { function Open -> P.OCM | Close -> P.CCM }
  | "y>"        { function Open -> P.OCY | Close -> P.CCY }
  | "k>"        { function Open -> P.OCK | Close -> P.CCK }
  | "w>"        { function Open -> P.OCW | Close -> P.CCW }
  | _           { function
                    | Open  -> error ("illegal tag: <"  ^ get lexbuf) 
                    | Close -> error ("illegal tag: </" ^ get lexbuf) 
                }    
{


let tok2str  = function
  | P.OB -> "<B>"
  | P.OEM -> "<EM>" 
  | P.OI -> "<I>"
  | P.OPL -> "<PL>"
  | P.OS -> "<S>"
  | P.OTT -> "<TT>"
  | P.OU -> "<U>"
  | P.OS0 -> "<0>"
  | P.OS1 -> "<1>"
  | P.OS2 -> "<2>"
  | P.OS3 -> "<3>"
  | P.OS4 -> "<4>"
  | P.OS5 -> "<5>"
  | P.OS6 -> "<6>"
  | P.OS7 -> "<7>"
  | P.OS8 -> "<8>"
  | P.OS9 -> "<9>"
  | P.OCR -> "<R>"
  | P.OCG -> "<G>"
  | P.OCB -> "<B>"
  | P.OCC -> "<C>"
  | P.OCM -> "<M>"
  | P.OCY -> "<Y>"
  | P.OCK -> "<K>"
  | P.OCW -> "<W>"
  | P.CB -> "</B>"
  | P.CEM -> "</EM>"
  | P.CI -> "</I>"
  | P.CPL -> "</PL>"
  | P.CS -> "</S>"
  | P.CTT -> "</TT>"
  | P.CU -> "</U>"
  | P.CS0 -> "</0>"
  | P.CS1 -> "</1>"
  | P.CS2 -> "</2>"
  | P.CS3 -> "</3>"
  | P.CS4 -> "</4>"
  | P.CS5 -> "</5>"
  | P.CS6 -> "</6>"
  | P.CS7 -> "</7>"
  | P.CS8 -> "</8>"
  | P.CS9 -> "</9>"
  | P.CCR -> "</R>"
  | P.CCG -> "</G>"
  | P.CCB -> "</B>"
  | P.CCC -> "</C>"
  | P.CCM -> "</M>"
  | P.CCY -> "</Y>"
  | P.CCK -> "</K>"
  | P.CCW -> "</W>"
  | P.EOF -> "<eof>"
  | P.TEXT(s)  -> ("TEXT:"^s)
  | P.SPACE(s) -> ("SPACE:"^ String.escaped s)

}
